﻿/*
 * SAT2MD_V3.c
 *
 * Created: 16.08.2014 18:04:36
 * Author: Jakob Schäfer
 */ 

#define F_CPU 20000000
#include <avr/io.h>
#include <avr/delay.h>
#include <avr/interrupt.h>

#define LedOn() {PORTB |= (1<<0); DDRB |= (1<<0);}
#define LedOff() { DDRB &= ~(1<<0); PORTB &= ~(1<<0);}
	
#define SAT_UP 0
#define SAT_DOWN 1
#define SAT_LEFT 2
#define SAT_RIGHT 3
#define SAT_START 4
#define SAT_A 5
#define SAT_B 6
#define SAT_C 7
#define SAT_X 8
#define SAT_Y 9
#define SAT_Z 10
#define SAT_L 11
#define SAT_R 12
#define PADTYPE_NC 0
#define PADTYPE_STD 1
#define PADTYPE_3D 2

void ReadSaturnController(void);
uint8_t WaitForTL(uint8_t val);
void PrepareMDRegisters(void);



uint8_t volatile megaDriveArray[8];
#define nextMD GPIOR0
#define nextIndex GPIOR1
#define indexMask GPIOR2
	
uint16_t volatile satButtons;
uint8_t satPadType = PADTYPE_NC;

int main(void)
{
	// config Saturn pins
	PORTD |= (1<<0)|(1<<1);
	DDRD |= (1<<0)|(1<<1);
	PORTD |= (1<<3)|(1<<4)|(1<<5)|(1<<6)|(1<<7);
	
	// config Megadrive pins
	DDRC |= (1<<0)|(1<<1)|(1<<2)|(1<<3)|(1<<4)|(1<<5);
	
	// enable timer2 overflow interrupt
	TIMSK2 |= (1<<TOIE2);
	
	// prepare MD data
	indexMask = 1;
	satButtons = 0xFFFF;
	PrepareMDRegisters();
		
	// enable INTO; falling & rising edge
	EICRA |= (1<<ISC00);
	EIMSK |= (1<<INT0);
	// enable interrupts now	
	sei();	
	
	// main loop
    while(1)
    {			
			
		ReadSaturnController();
		if (satPadType != PADTYPE_NC){
			PrepareMDRegisters();
			LedOn();	
			_delay_ms(5);
		}
		else{
			LedOff();
		}	
		
    }
}

 ISR(INT0_vect, ISR_NAKED){

	asm volatile (	
		// output nextMD on PORTC
		"	in __zero_reg__, 0x1E						\n"
		"	out 0x08, __zero_reg__						\n"
		
		// save registers
		"	push r16									\n"
		"	push r30									\n"
		"	push r31									\n"
		"	in r16, __SREG__							\n"	
		"	push r16									\n"		
		
		// start timer2 -> overflow after 1.64 ms	
		"	ldi r16, 5									\n"		
		"	sts 0xB1, r16								\n"
		
		// nextIndex++
		"	in r16, 0x2A								\n"	
		"	inc r16										\n"	
		// apply index mask (1 or 7)
		"	in __zero_reg__, 0x2B						\n"			
		"	and r16, __zero_reg__						\n"	
		"	out 0x2A, r16								\n"	
		
		// array address in z pointer
		"	ldi r30, lo8(megaDriveArray)				\n"	
		"	ldi r31, hi8(megaDriveArray)				\n"	
		"	add r30, r16								\n"
		"	clr r16										\n"
		"	adc r31, r16								\n"
		
		// load next value into nextMD
		"	ld r16, Z												\n"
		"	out 0x1E, r16											\n"				
		
		// restore registers
		"	clr __zero_reg__						\n"
		"	pop r16									\n"		
		"	out __SREG__, r16						\n"		
		"	pop r31									\n"
		"	pop r30									\n"
		"	pop r16									\n"
		"	reti									\n"
	::);
}

 ISR(TIMER2_OVF_vect){
	 
	// stop and reset timer2
	TCCR2B = 0x00;
	TCNT2 = 0x00;
	 	 
	// check select and output standard data
	if (PIND & (1<<2)){
		PORTC = megaDriveArray[0];
		nextMD = megaDriveArray[1];
		nextIndex = 1;
	}
	else{
		PORTC = megaDriveArray[1];
		nextMD = megaDriveArray[2];
		nextIndex = 2;
	} 	 
	
 }


void ReadSaturnController(void){
	
	#define ResetTH() {PORTD &= ~(1<<0);}
	#define SetTH() {PORTD |= (1<<0);}
	#define ResetTR() {PORTD &= ~(1<<1);}	
	#define SetTR() {PORTD |= (1<<1);}
		
	uint16_t temp = 0xFFFF;
	uint8_t state = (PIND & 0b10111000);
		
	// what kind of controller is plugged in?
	switch (state){
		
		// standard controller:
		case 0b10100000:
		
			if ( !(PIND & (1<<6)) ) temp &= ~(1<<SAT_L);
			
			ResetTH();
			_delay_us(5);			
			if ( !(PIND & (1<<3)) ) temp &= ~(1<<SAT_UP);
			if ( !(PIND & (1<<4)) ) temp &= ~(1<<SAT_DOWN);
			if ( !(PIND & (1<<5)) ) temp &= ~(1<<SAT_LEFT);
			if ( !(PIND & (1<<6)) ) temp &= ~(1<<SAT_RIGHT);
			
			SetTH();
			ResetTR();
			_delay_us(5);
			if ( !(PIND & (1<<3)) ) temp &= ~(1<<SAT_B);
			if ( !(PIND & (1<<4)) ) temp &= ~(1<<SAT_C);
			if ( !(PIND & (1<<5)) ) temp &= ~(1<<SAT_A);
			if ( !(PIND & (1<<6)) ) temp &= ~(1<<SAT_START);
			
			ResetTH();
			_delay_us(5);
			if ( !(PIND & (1<<3)) ) temp &= ~(1<<SAT_Z);
			if ( !(PIND & (1<<4)) ) temp &= ~(1<<SAT_Y);
			if ( !(PIND & (1<<5)) ) temp &= ~(1<<SAT_X);
			if ( !(PIND & (1<<6)) ) temp &= ~(1<<SAT_R);
			
			SetTH();
			SetTR();
			
			// activate three button mode if R trigger is pushed
			if (satPadType == PADTYPE_NC){
				if (!(temp&(1<<SAT_R))){
					indexMask = 1;		
				}
				else{
					indexMask = 7;	
				}
			}	
			
			satPadType = PADTYPE_STD;
			break;		
			
		// 3D controller
		case 0b10001000:
			
			
			ResetTH();
			_delay_us(50);
		
			ResetTR();		
			if (WaitForTL(0) == 0) break;
			_delay_us(10);
			
			SetTR();
			if (WaitForTL(1) == 0) break;
			_delay_us(10);
			
			ResetTR(); 
			if (WaitForTL(0) == 0) break;
			_delay_us(10);
			if ( !(PIND & (1<<3)) ) temp &= ~(1<<SAT_UP);
			if ( !(PIND & (1<<4)) ) temp &= ~(1<<SAT_DOWN);
			if ( !(PIND & (1<<5)) ) temp &= ~(1<<SAT_LEFT);
			if ( !(PIND & (1<<6)) ) temp &= ~(1<<SAT_RIGHT);
			
			
			SetTR();
			if (WaitForTL(1) == 0) break;
			_delay_us(10);
			if ( !(PIND & (1<<3)) ) temp &= ~(1<<SAT_B);
			if ( !(PIND & (1<<4)) ) temp &= ~(1<<SAT_C);
			if ( !(PIND & (1<<5)) ) temp &= ~(1<<SAT_A);
			if ( !(PIND & (1<<6)) ) temp &= ~(1<<SAT_START);
			
			
			ResetTR();		
			if (WaitForTL(0) == 0) break;
			_delay_us(10);
			if ( !(PIND & (1<<3)) ) temp &= ~(1<<SAT_Z);
			if ( !(PIND & (1<<4)) ) temp &= ~(1<<SAT_Y);
			if ( !(PIND & (1<<5)) ) temp &= ~(1<<SAT_X);
			if ( !(PIND & (1<<6)) ) temp &= ~(1<<SAT_R);
			
			SetTR();
			if (WaitForTL(1) == 0) break;
			_delay_us(10);
			if ( !(PIND & (1<<6)) ) temp &= ~(1<<SAT_L);
			
			ResetTR();		
			if (WaitForTL(0) == 0) break;
			_delay_us(10);
			
			SetTR();
			if (WaitForTL(1) == 0) break;
			_delay_us(10);
			
			// end
			_delay_us(100);
			SetTH();
			
			// activate three button mode if R trigger is pushed
			if (satPadType == PADTYPE_NC){
				if (!(temp&(1<<SAT_R))){
					indexMask = 1;		
				}
				else{
					indexMask = 7;	
				}
			}
			
			satPadType = PADTYPE_3D;
			break;
		
		// no Saturn controller attached	
		default:
			satPadType = PADTYPE_NC;
			break;			
	}

	satButtons = temp;
}

uint8_t WaitForTL(uint8_t val){
	
	// result = 1 => ok; result = 0 => timer overflow
	uint8_t result;	
	
	// start timer0 with prescaler = 256 => timer0 overflow after 3 ms
	TCNT0 = 0x00;
	TCCR0B |= (1<<CS02); 
	
	// wait until TL (PIN D7) is low...
	if (val == 0x00){
		while (  (PIND&(1<<7)) && !(TIFR0&(1<<TOV0))  ); 
	}
	// or wait until TL is high
	else{
		while (  !(PIND&(1<<7)) && !(TIFR0&(1<<TOV0))  );	
	}
	
	// if timer0 hasn't overflowed, everything's ok
	TCCR0B &= ~(1<<CS02);
	if ( (TIFR0 & (1<<TOV0)) == 0 ){
		result = 1;
	}
	else{
		TIFR0 |= (1<<TOV0);
		result = 0;
	}
	
	return result;
}

void PrepareMDRegisters(void){
	
	uint8_t selectLow = 0xFF;
	uint8_t selectHigh = 0xFF;
	uint8_t selectXYZ = 0xFF;
	
	if ( !(satButtons & (1<<SAT_UP)) )		{	selectLow &= ~(1<<0);	selectHigh &= ~(1<<0);	}
	if ( !(satButtons & (1<<SAT_DOWN)) )	{	selectLow &= ~(1<<1);	selectHigh &= ~(1<<1);	}
	if ( !(satButtons & (1<<SAT_LEFT)) )	{	selectHigh &= ~(1<<2);		}
	if ( !(satButtons & (1<<SAT_RIGHT)) )	{	selectHigh &= ~(1<<3);		}
	if ( !(satButtons & (1<<SAT_A)) )		{	selectLow &= ~(1<<4);	}
	if ( !(satButtons & (1<<SAT_B)) )		{	selectHigh &= ~(1<<4);		}
	if ( !(satButtons & (1<<SAT_C)) )		{	selectHigh &= ~(1<<5);		}
	if ( !(satButtons & (1<<SAT_START)) )	{	selectLow &= ~(1<<5);	}
	if ( !(satButtons & (1<<SAT_X)) )		{	selectXYZ &= ~(1<<2);	}
	if ( !(satButtons & (1<<SAT_Y)) )		{	selectXYZ &= ~(1<<1);	}
	if ( !(satButtons & (1<<SAT_Z)) )		{	selectXYZ &= ~(1<<0);	}
	if ( !(satButtons & (1<<SAT_R)) )		{	selectXYZ &= ~(1<<3);	}
		
	megaDriveArray[0] = selectHigh;
	megaDriveArray[1] = selectLow & 0xF3;
	megaDriveArray[2] = selectHigh;
	megaDriveArray[3] = selectLow & 0xF3;
	megaDriveArray[4] = selectHigh;
	megaDriveArray[5] = selectLow & 0xF0;
	megaDriveArray[6] = selectXYZ;
	megaDriveArray[7] = selectLow | 0x0F;
	
	// output new data if timer2 is not running (= Megadrive is not polling atm)
	if (TCCR2B == 0){
		if (PIND & (1<<2)){
			PORTC = megaDriveArray[0];
			nextMD = megaDriveArray[1];
			nextIndex = 1;
		}
		else{
			PORTC = megaDriveArray[1];
			nextMD = megaDriveArray[2];
			nextIndex = 2;
		} 	 
	}		
}	